-- A cycle non-termination checker, which can also prove relative cycle nontermination

module Main where
import CYCSRS.Processor
import CYCSRS.Types
import CYCSRS.Util.System
import Data.Algorithms.KMP
import System.IO
import System.Environment
import Data.IORef
import Data.List
import Data.Maybe
import Data.Function
import Control.Concurrent
import qualified System.Timeout as ST
import qualified Data.Set as Set
import GHC.Conc (numCapabilities)
import Debug.Trace

s `isInfix` t = not $ null $ findMatches s t
findMatches pat string = match (build pat) string

rewrite numbers rules string  =
 concatMap (\r -> rewriteOneRule r string) (zip rules numbers)
 
-- rewriteOneRule :: (Rule,Int) -> Sequence -> (Sequence,Int) 
rewriteOneRule (rule@(left,right), rulenumber) string =
 let positions = findMatches left string
     l = length left
 in [(replace pos l right string,rulenumber) | pos <- positions ]

replace pos removelength new string =
 let (prefix,suffix) = splitAt pos string 
 in prefix ++ new ++ (drop removelength suffix)

findNontermOneString :: SRS -> SRS -> Sequence -> Int -> ProofResult [(Sequence, Int, Int)]
findNontermOneString weaksrs strictsrs string depthbound = 
 case dfsPS goal succ stack depthbound applyToResult of
  Nothing -> None
  Just x  -> x
 where 
   goal :: [(Sequence, Int, Int)] -> Bool
   goal q@((x,a,_):rest)
   -- case that we prove non-termination but not relative non-termination,
   -- i.e. there are no weak rules (all rules are strict)
    | null weaksrs && a /= 0 && string `isInfix` x = True
   -- relative termination: the looping path must contain at least one
   -- step where a strict rule was applied 
    | a /= 0 
      && string `isInfix` x 
      && not (null [num | (_,num,_) <- q, num > 0]) = True
    | otherwise = False
   succ = \string -> (rewrite [1..] strictsrs string) ++ (rewrite [-1,-2..] weaksrs string)
   stack = [(string,0)]
   applyToResult p = let (rx@(x,_,_):xs) = reverse p
                         (r,(a,b,c):_)   = break (\(y,_,_) -> y `isInfix` x) xs
                     in StringExpansion ((a,0,c):(reverse (rx:r))) 




   
-- u^a v^b --> v^c u^d                                                  
   
isCyclicExpansion w1 w2 = isJust (cyclicExpansion w1 w2) 

cyclicExpansion w1 w2 =
   if null interpretW1 then Nothing 
    else Just $ head $ interpretW1
 where
  interpretW2  =
     let l = length w2
     in
     [(u,v,a,b) | u <- (take (l) $ tail $ inits w2), 
              v <- (tails w2),
              length u + length v <= l,
              (a,b) <- [(x,y) | x <- [1..l], y <- [0..l-x] ],
              u ^^^ a ++ v ^^^ b == w2,
              if null v then b == 0 else True ]
              
  interpretW1 =
     let l = length w1
     in
     [(u,v,a,b,c,d) | (u,v,a,b) <- interpretW2,
                  (c,d) <- [(x,y) | x <- [0..l], y <- [1..l-x] ],
                   v ^^^ c ++ u ^^^ d  == w1,
                   if null v then c == 0 else True,
                   c >= b, d >= a, c == b || d == a]
       
  u ^^^ n  = concat $ replicate n u             
                                                                                                                                                                 


merge4 (x1:xs1) (x2:xs2) (x3:xs3) (x4:xs4) = x1:x2:x3:x4:(merge4 xs1 xs2 xs3 xs4)
merge4 []   xs2 xs3 xs4 = merge3 xs2 xs3 xs4
merge4 xs2  [] xs3 xs4  = merge3 xs2 xs3 xs4
merge4 xs2 xs3 [] xs4   = merge3 xs2 xs3 xs4
merge4 xs2 xs3 xs4 []   = merge3 xs2 xs3 xs4


merge3 (x1:xs1) (x2:xs2) (x3:xs3) = x1:x2:x3:(merge3 xs1 xs2 xs3)
merge3 [] xs2 xs3 = merge2 xs2 xs3
merge3 xs2 [] xs3 = merge2 xs2 xs3
merge3 xs2 xs3 [] = merge2 xs2 xs3

merge2 (x1:xs1) (x2:xs2) = x1:x2:(merge2 xs1 xs2)
merge2 [] xs = xs
merge2 xs [] = xs
merge (x1:x2:x3:x4:x5:xs) (y:ys) = x1:x2:x3:x4:x5:y:(merge xs ys)
merge (x:xs) (y:ys) = x:y:(merge xs ys)
merge [] ys = ys
merge xs [] =xs                                                                                                                                                                    
-- searchNontermHeur :: SRS -> String  
searchNontermHeur weaksrs strictsrs = 
 let srs = weaksrs ++ strictsrs
     lhss = map fst srs
     rhss = map snd srs
     alpha = nub $ (concat lhss) ++ (concat rhss)
     mx = maximum $ map length lhss
     mxx = max mx (maximum $ map length rhss)     
     extensions n = [lhs ++ y |  lhs <- lhss, y <- (gen alpha n)]
     extensionsS  0 = lhss
     extensionsS  1 = concat [ [(y:str), (str++[y])] | str <- (sortBy (compare `on` length) lhss),  y <- alpha]   
     extensionsS  n = 
        let rek = extensionsS (n-1)
            n1 = [(y:str) | str <- rek, y <- alpha]
            n2 = [str ++ [y] | str <- rek, y <-alpha]
        in Set.elems $ Set.union (Set.fromList n1) (Set.fromList n2)
        
     problemset1 =  
        [ findNontermOneStringSRSNT weaksrs strictsrs string bound 
          | (i,bound) <- [(k,5) | k <- [0..4*mxx]] -- ..mxx
          , string <- extensionsS i
        ]                      
     problemset3 = 
        [ findNontermOneString weaksrs strictsrs string bound 
          | (i,bound) <- [(k,30) | k <- [0..max 5 mxx]] -- max 5
          , string <- extensionsS i
        ]
     problemset4 = 
        [ findNontermOneCycle weaksrs strictsrs string bound 
           | (i,bound) <- [(k,30) | k <- [0..max 5 mxx]] -- max 5
           ,string <- extensions i
        ]
     problemset2 = 
        [ findNontermOneStringSRSNT weaksrs strictsrs string bound 
           | (i,bound) 
                <- [(k,30) | k <- [0..max 5 mxx]] -- max 5
                    ++
                   [(k,10) | k <- [(max 5 mxx)+1..2*mxx]] -- max 5 .. (k,10)
                    ++
                   [(k,5) | k <- [2*mxx+1..3*mxx]]
                     ++
                   [(k,50) | k <- [0..10*mxx]]
           , string <- extensionsS i]

     simplePar mvar pbs = case filter isNotNone pbs of
                            (x:xs) -> putMVar mvar (fromListRes x weaksrs strictsrs)
                            []     -> putMVar mvar (MAYBE,"")
  in
   do
    res <- newEmptyMVar
    forkIO (simplePar res problemset1)
    forkIO (simplePar res problemset2)
    forkIO (simplePar res problemset3)
    forkIO (simplePar res problemset4)
    (a1,b1) <- takeMVar res
    case a1 of
     NO ->    return b1
     YES ->   return b1
     other -> do
                (a2,b2) <- takeMVar res
                case a2 of
                  NO ->    return b2
                  YES ->   return b2
                  other -> do
                            (a3,b3) <- takeMVar res
                            case a3 of
                              NO ->    return b3
                              YES ->   return b3
                              other -> do 
                                        (a4,b4) <- takeMVar res
                                        return b4
    
fromListRes :: (ProofResult [(Sequence, Int, Int)]) -> SRS -> SRS -> (Result,String)
fromListRes list weaksrs strictsrs 
 -- we prove cycle-non-termination, i.e. there are no weak rules:
 | null weaksrs = 
    let srs = weaksrs ++ strictsrs 
    in case list of 
        None              -> (MAYBE,"")
        CyclicExpansion x -> 
          let proof = map (\(a,b,c) -> (a,b)) $  x
              w2     = (\(a,b) -> a) $  head proof 
              w1     = (\(a,b) -> a) $  last proof                      
              (u,v,a,b,c,d) = fromJust $ cyclicExpansion w1 w2
          in (NO,
                unlines 
                    ["NO"
                    ,"The SRS: "
                    ,showSRS srs 
                    ,"is cycle-nonterminating."
                    ,"Proof: "
                    ,"There exists a string derivation from w = u^" ++ show a ++ "v^" ++ show b ++ " to v^" ++ show c ++ "u^" ++show d  ++ " = w'"
                    ,"u      = " ++ (concat $ intersperse " " u)
                    ,"v      = " ++ (concat $ intersperse " " v)
                    ,"a      = " ++ show a
                    ,"b      = " ++ show b
                    ,"w      = " ++ (concat $ intersperse " " w2)
                    ,"i.e w  = (" ++ (concat $ intersperse " " u) ++ ")^"++ show a ++ "(" ++ (concat $ intersperse " " v) ++")^" ++ show b 
                    ,"c      = " ++ show c
                    ,"d      = " ++ show d
                    ,"w'     = " ++ (concat $ intersperse " " w1)
                    ,"i.e w' = (" ++ (concat $ intersperse " " v) ++ ")^"++ show c ++ "(" ++ (concat $ intersperse " " u) ++")^" ++ show d 
                    ,""
                    ,"       The string derivation is:"                  
                    ,""
                    ,showProofSRS weaksrs strictsrs proof (concat $ intersperse " " w2) 
                    ,""
                    ,"Thus the system is cycle-nonterminating.               QED"
                  ])
        CycleLoop x -> (NO,
                 let proof = map (\(a,b,c) -> (a,b)) $ x
                     w     = (\(a,b) -> a) $  head proof       
                 in
                 unlines 
                  ["NO"
                  ,"The SRS: "
                  ,showSRS srs 
                  ,"is cycle-nonterminating."
                  ,"Proof: There exists a looping derivation [w] o->^+ [w] where"
                  ,"         [w] = " ++ (concat $ intersperse " " w)
                  ,""
                  ,"       The derivation is:"                  
                  ,""
                  ,showProofSRS weaksrs strictsrs proof (concat $ intersperse " " w) 
                  ,""
                  ,"Thus the system is cycle-nonterminating.               QED"
                  ])
        StringExpansion x -> (NO,
                 let proof = map (\(a,b,c) -> (a,b)) $ x
                     w     = (\(a,b) -> a) $  head proof       
                 in
                 unlines 
                  ["NO"
                  ,"The SRS: "
                  ,showSRS srs 
                  ,"is string-nonterminating."
                  ,"Proof: There exists a  string derivation w ->^+ uwv where"
                  ,"         w = " ++ (concat $ intersperse " " w)
                  ,""
                  ,"       The derivation is:"                  
                  ,""
                  ,showProofSRS weaksrs strictsrs proof (concat $ intersperse " " w) 
                  ,""
                  ,"Thus the system is string-nonterminating. "
                  ,"String-nontermination implies cycle-nontermination."
                  ,"Thus the system is cycle-nonterminating.            QED"

                  ])
 
 -- there are weak rules, i.e. we prove relative cycle-non-termination:
 | otherwise =
    case list of 
       (CyclicExpansion x) -> 
                 let proof = map (\(a,b,c) -> (a,b)) $ x
                     w2     = (\(a,b) -> a) $  head proof 
                     w1     = (\(a,b) -> a) $  last proof                      
                     (u,v,a,b,c,d) = fromJust $ cyclicExpansion w1 w2
                 in (NO,
                  unlines 
                    ["NO"
                    ,"The SRS: "
                    ,showSRSRel (weaksrs,strictsrs) 
                    ,"is relatively cycle-nonterminating."
                    ,"Proof: There exists a string derivation from w = u^" ++ show a ++ "v^" ++ show b ++ " to v^" ++ show c ++ "u^" ++show d  ++ " = w'"
                    ,"u      = " ++ (concat $ intersperse " " u)
                    ,"v      = " ++ (concat $ intersperse " " v)
                    ,"a      = " ++ show a
                    ,"b      = " ++ show b
                    ,"w      = " ++ (concat $ intersperse " " w2)
                    ,"i.e w  = (" ++ (concat $ intersperse " " u) ++ ")^"++ show a ++ "(" ++ (concat $ intersperse " " v) ++")^" ++ show b 
                    ,"c      = " ++ show c
                    ,"d      = " ++ show d
                    ,"w'     = " ++ (concat $ intersperse " " w1)
                    ,"i.e w' = (" ++ (concat $ intersperse " " v) ++ ")^"++ show c ++ "(" ++ (concat $ intersperse " " u) ++")^" ++ show d 
                    ,""
                    ,"       The string derivation using at least one strict rule is:"                  
                    ,""
                    ,showProofSRS weaksrs strictsrs proof (concat $ intersperse " " w2) 
                    ,""
                    ,"Thus the system is relatively cycle-nonterminating.               QED"
                  ])
       (CycleLoop x) -> (NO,
                 let proof = map (\(a,b,c) -> (a,b)) $  x
                     w     = (\(a,b) -> a) $  head proof       
                 in unlines
                  ["NO"
                  ,"The relative SRS: "
                  ,showSRSRel (weaksrs,strictsrs) 
                  ,"is relatively cycle-nonterminating."
                  ,"Proof: There exists a looping derivation [w] o->^+ [w] where"
                  ,"         [w] = " ++ (concat $ intersperse " " w)
                  ,""
                  ,"       The derivation is:"                  
                  ,""
                  ,showProof weaksrs strictsrs proof (concat $ intersperse " " w) 
                  ,""
                  ,"Thus the system is relatively cycle-nonterminating.               QED"
                  ])
       (StringExpansion x) -> (NO,
                 let proof = map (\(a,b,c) -> (a,b)) $ x
                     w     = (\(a,b) -> a) $  head proof       
                 in
                 unlines 
                  ["NO"
                  ,"The relative SRS: "
                  ,showSRSRel (weaksrs,strictsrs) 
                  ,"is relatively string-nonterminating."
                  ,"Proof: There exists a  string derivation w ->^+ uwv where"
                  ,"         w = " ++ (concat $ intersperse " " w)
                  ,"       and at least one step is by a strict rule."
                  ,"       The string derivation is:"                  
                  ,""
                  ,showProofSRS weaksrs strictsrs proof (concat $ intersperse " " w) 
                  ,""
                  ,"Thus the system is relatively string-nonterminating. "
                  ,"Relative string-nontermination implies relative cycle-nontermination."
                  ,"Thus the system is relatively cycle-nonterminating.            QED"
                  ])
       _  -> (MAYBE,"")

showProofSRS weaksrs strictsrs proof w = let (x:xs) = go 7 proof in x ++ (unlines xs)
 where
  strict_string_rules = (map showRule strictsrs)
  weak_string_rules = (map showRuleRel weaksrs)
  string_rules = strict_string_rules ++ weak_string_rules
  offset2     = 10 + (maximum $ map length string_rules)
  go _ []  = [] 
  go u ((s,0):xs) = (replicate u ' ' ++ (concat $ intersperse " " s)):(go 1 xs) 
  go u ((s,i):xs) = let rule = " ->" ++ (if i < 0 then "=_{" ++ weak_string_rules!!((abs i)-1) else "_{" ++ strict_string_rules!!(i-1)) ++ " }  " 
                        rule' = rule ++ (replicate (offset2 - length rule) ' ')
                    in (replicate (u) ' ' ++ rule'  ++ (concat $ intersperse " " s)):(go (8+offset) xs) 
  offset = length $ concat $ intersperse " " $ fst $ head proof                 
 
cyclerewrite numbers rules string =
 concatMap (\r -> cyclerewriteOneRule r string) (zip rules numbers)

cyclerewriteOneRule (rule@(left,right), rulenumber) string =
 let positions = if l > (length string) then [] else findMatches left (string++string)
     l = length left
 in [(cyclereplace pos l right string,rulenumber) | pos <- positions, pos < length string ]
 
cyclereplace pos removelength new string =
 let res1 = replace pos removelength new (string++string)
     offset = max 0 (removelength - (length string - pos))
 in take (length string - removelength + length new) (drop offset res1)
 
 
isSim w1 w2 
 | length w1 == length w2 = not (null (findMatches w1 (w2++w2)))
 | otherwise = False

data ProofResult a = None | CycleLoop a | CyclicExpansion a  | StringExpansion a
 deriving(Show)

isNotNone None = False
isNotNone _    = True

fromCycleLoop (CycleLoop  a) = a
fromCyclicExpansion (CyclicExpansion a ) = a
 
findNontermOneStringSRSNT :: SRS -> SRS -> Sequence -> Int -> ProofResult [(Sequence, Int, Int)]
findNontermOneStringSRSNT weaksrs strictsrs string depthbound = 
   case dfsPS goal succ stack depthbound applyToResult of
    Just a ->  a
    Nothing -> None
 where 
   goal p@((x,a,_):rest) 
   -- no weak rules:
     | a /= 0 && null weaksrs && string `isInfix`  x  =  True -- any (== True) (map (\(string,_,_) -> string `isInfix` x) rest) = True
     | a /= 0 && null weaksrs && x `isCyclicExpansion` string = True -- any (== True) (map (\(string,_,_) -> x `isCyclicExpansion` string) rest) = True
  -- with weak rules
     | a /= 0 && not (null [a | (_,a,_) <- p, a > 0]) && string `isInfix`  x  =  True -- any (== True) (map (\(string,_,_) -> string `isInfix` x) rest) = True
     | a /= 0 && not (null [a | (_,a,_) <- p, a > 0]) && x `isCyclicExpansion` string =  True -- any (== True) (map (\(string,_,_) -> x `isCyclicExpansion` string) rest) = True
   goal _ = False  
     
   applyToResult p = let (rx@(x,_,xx):xs) = reverse p
                         (ry@(u,_,uu):_) = p
                         stringcheck = u `isInfix` x
                         chk = null [u | (_,u,_) <- p]
                     in -- trace (show p)  $
                         if stringcheck then 
                           StringExpansion p
                         else
                           CyclicExpansion p
                     
   succ = \string -> (rewrite [1..] strictsrs string) ++ (rewrite [-1,-2..] weaksrs string)
   stack = [(string,0)]

   
findNontermOneCycle :: SRS -> SRS -> Sequence -> Int -> ProofResult [(Sequence, Int, Int)]
findNontermOneCycle weaksrs strictsrs string depthbound = 
  
   case dfsPS goal succ stack depthbound applyToResult of
    Just a ->  a
    Nothing -> None
 where
   goal q@((x,a,_):rest) 
    | null weaksrs &&  a /= 0 && string `isSim` x = True
    | not (null [w | (_,w,_) <- q, w > 0]) &&  a /= 0 && string `isSim` x = True
    | otherwise = False
  
   applyToResult p = let (rx@(x,_,_):xs) = reverse p
                         (r,(a,b,c):_) = break (\(y,_,_) -> isSim x y) xs

                     in CycleLoop ((a,0,c):(reverse (rx:r))) 
   succ = \string -> (cyclerewrite [1..] strictsrs string) ++ (cyclerewrite [-1,-2..] weaksrs string)
   stack = [(string,0)]
showProof weaksrs strictsrs proof w = 
  let (x:xs) = go 7 proof 
  in
   (x ++ (unlines xs))
 where
  strict_string_rules = (map showRule strictsrs)
  weak_string_rules = (map showRuleRel weaksrs)
  string_rules = strict_string_rules ++ weak_string_rules
  offset2     = 11 + (maximum $ map length string_rules)
  go _ []  = [] 
  go u ((s,0):xs) = (replicate u ' ' ++ (concat $ intersperse " " s)):(go 1 xs) 
  go u ((s,i):xs) = let rule = " o->" ++ (if i < 0 then "=_{" ++ weak_string_rules!!((abs i)-1) else "_{" ++ strict_string_rules!!(i-1)) ++ " }  " 
                        rule' = rule ++ (replicate (offset2 - length rule) ' ')
                    in (replicate (u) ' ' ++ rule'  ++ (concat $ intersperse " " s)):(go (8+offset) xs) 
  offset = length $ concat $ intersperse " " $ fst $ head proof                 
gen alpha i 
 | i <= 0 =  [[]]     
gen alpha i = [a:xs | a <- alpha, xs <- (gen alpha (i-1))]     

{-     
main =
 do 
  [a,b] <- getArgs 
  (weaksrs,strictsrs) <- 
          if "xml" `isSuffixOf` a then
            getRelativeXML a
           else getSRSRel a
  -- let (r,p) = searchNontermHeur con  
  if null weaksrs then do
    p <- searchNontermHeur [] strictsrs  
    putStrLn p
   else do 
    p <- searchNontermHeur weaksrs strictsrs
    putStrLn p
-}        
  

   
check file =
 do 
  con <- getXML file
  return $ searchNontermHeur con
{-
processor = 
 Processor {
    properties = PProperties {cycletermination = False,
                              cyclenontermination=True,
                              transformator=False,
                              description="Brute-force search for cycle nontermination"},
    run = runNT}
  where
        runNT procinput 
         | Just _ <- weakproblem procinput = return 
                                            (POutput {proof =  "PROCESSOR: Cycnt\n skip - unable to prove relative termination"
                                                      ,result = MAYBE
                                                      ,outputproblem = inputproblem procinput
                                                      ,outputweakproblem = weakproblem procinput})

         | Nothing <- weakproblem procinput =
        
         do
           let o = options procinput
           let maybetimeout = timeout o
           let srs = inputproblem procinput
           let runIt =  return (searchNontermHeur srs)
           result <- case maybetimeout of
                       Nothing -> runIt >>= \r -> return (Just r)
                       Just t  -> ST.timeout (t*10^6) runIt
           case result of 
            Nothing -> return $ POutput {proof =  "PROCESSOR: Cycle-Nontermination-Checker\nCycle-Nontermination-Checker is applied to the problem\n"++showSRS srs ++ "\n\nOutput of Cycle-Nontermination-Checker:\n", result = MAYBE, outputproblem = srs, outputweakproblem=Nothing}
            Just (r,p) -> return $ POutput {proof = "PROCESSOR: Cycle-Nontermination-Checker\nCycle-Nontermination-Checker is applied to the problem\n"++showSRS srs ++ "\n\nOutput of Cycle-Nontermination-Checker:\n" ++ p, result = r, outputproblem =[],outputweakproblem=Nothing}            
-}
  
{-
allSim (list,len) =
 take len (go (cycle list))
 where
  go ys@(x:xs) = (take len ys):(go xs)
  
-- rewrite :: Ord a => [([a],[a])] -> CycList a -> [[Integer]]  
rewrite rules (list,len) = 
   let string = (take (2*len) (cycle list))
       alpha = list ++ (concatMap fst rules) ++ (concatMap snd rules)
       eof  = (head list) ++ (maximumBy (\a b -> compare (length a) (length b)) alpha)
       clean xs = map (\x -> x - 1) $ takeWhile (\x -> x <= fromIntegral len) $ sort xs
       patterns = map fst rules
       positions = map clean $ matches patterns string eof
   in positions
   
example = 
 let cycle = listToCycle (concat $ replicate 100 ["a","b"])
     rules = [(["a","b"] , ["b","a","b"])
             ,(["a","b","a"] , ["a","a","b"])     
             ,(["b","b","b"] , ["b","a","b"])]
 in rewrite rules cycle             

example1 = 
 let cycle = listToCycle ([[1],[2],[3],[4],[5]])
     rules = [([4,5] , [7,8,9])]
 in rewrite rules cycle             
 -}
-- replace pos lenl r (xs,len) =
 -- let 
  -- dropcount = lenl
  -- (before,after) = splitAt pos xs  
         
-- test =
 -- let 
  -- goal ((x,_):_) = x == 100
  -- succ x = [(2*x,"double"),(2*x+1,"ndouble")]
  -- start  = [(1,"start")]
  -- bound = 10
 -- in dfsP goal succ start bound 
   
-- searches for cyclic loop [w] o->^+ [w]      
{-
searchNonterm srs = 
 let lhss = map fst srs
     rhss = map snd srs
     alpha = nub $ (concat lhss) ++ (concat rhss)
     mx = maximum $ map length lhss
     extensions n = [lhs ++ y |  lhs <- lhss, y <- (gen alpha (mx+n-(length lhs)))]
 in  case filter (isNotNone) [findNontermOneCycle srs string bound | (i,bound) <- [(k,50) | k <- [0..]], string <- extensions i] of
       []  -> "MAYBE"
       (CycleLoop x:xs) -> 
                 let proof = map (\(a,b,c) -> (a,b)) $ head $ x
                     w     = (\(a,b) -> a) $  head proof       
                 in
                 unlines 
                  ["NO"
                  ,"The SRS: "
                  ,showSRS srs 
                  ,"is cycle-nonterminating."
                  ,"Proof: There exists a looping derivation [w] o->^+ [w] where"
                  ,"         [w] = " ++ (concat $ intersperse " " w)
                  ,""
                  ,"       The derivation is:"                  
                  ,""
                  ,showProof srs proof (concat $ intersperse " " w) 
                  ,""
                  ,"Thus the system is cycle-nonterminating.               QED"
                  ]
-}                  

-- dfsP :: (a -> Bool) -> (a -> [a]) -> [[a]]   
dfsP goal succ stack depthbound applyToResult = 
  go [(map (\(a,c) -> (a,c,depthbound)) stack)]
 where
  -- go s 
   -- | trace (show s) False = undefined
  go [[]] = Nothing
  go []   = Nothing
  go ([]:(a:as):xs) = go (as:xs)
  go (((_,attribute,0):_):(a:as):xs) = go (as:xs)
  go l@(((k,attribute,b):r):xs) 
   | goal (map head l)    = Just [applyToResult (reverse $  map head l)]
   | otherwise = let successors = map (\(a,att) -> (a,att,b-1)) (succ k)
                 in go (successors:l) 
                  
   
  go x = error $ "in go " ++ show x
     


dfsPS goal succ stack depthbound applyToResult = 
  go [(map (\(a,c) -> (a,c,depthbound)) stack)] (Set.empty)
 where
  initnodes = Set.fromList (map fst stack)
--   go s vis
--      | trace (show $ s) False = undefined
  go [[]] vis = Nothing
  go []   vis =  Nothing
  go ([]:(a:as):xs) vis = go (as:xs) vis
  go (((_,attribute,0):_):(a:as):xs) vis = go (as:xs) vis
  go l@(((k,attribute,b):r):xs) vis
   | goal (map head l)    = Just $ applyToResult (reverse $  map head l)
   | otherwise = let 
                     succNub    = [(a,att) | (a,att) <- (succ k), not (a `Set.member` vis) || (a `Set.member` initnodes)]
                     successors = map (\(a,att) -> (a,att,b-1)) (succNub)
                 in go (successors:l)  (Set.insert k vis)
                 
isExp u v a b  str =                  
       let                  
           patu = (build $ reverse u)
           patv = (build v)
           lu = length u
           lv = length v

           len = length str
           rstr = reverse str
           takeW [] i j = []
           takeW (x:xs) i j 
             | i==x = (x:(takeW xs (i+j) j)) 
             | otherwise = []

           umatches =  case takeW [a | a <- match (patu) rstr, a `mod` lu == 0] 0 lu 
                         of 
                          [] -> Nothing
                          (x:xs) -> if x == 0 then Just (reverse (zip [1..] [len-a-(lu) | a    <- (x:xs)])) else Nothing
           vmatches = case takeW [a | a <- match (patv) str, a `mod` (lv) == 0] 0 lv of
                        [] -> Nothing
                        (x:xs) -> if x== 0 then Just (zip [1..] (x:xs)) else Nothing
           ustart = case vmatches of 
                      Just vss -> Just [ (a+ (length v),i) | (i,a) <- vss   ]
                      Nothing -> Nothing
           us = case umatches of        
                 Nothing -> Nothing
                 Just uss -> Just $ uss     
           res = case ustart of 
                      Just ws -> case us of
                                     Just zs -> [(fromJust x,i) | (i,a) <- zs, let x= (lookup a  ws), isJust x]
                                     Nothing -> []

       in  (ustart,res,umatches,vmatches, [() | (d,c) <- res, c >= a, d >= b, c == a || d == b ])
       
main =
 do 
  [a,b] <- getArgs 
  ST.timeout (10^6*read b) (
   do
      (weaksrs,strictsrs) <- 
              if "xml" `isSuffixOf` a then
                getRelativeXML a
               else getSRSRel a
      -- let (r,p) = searchNontermHeur con  
      p <- searchNontermHeur  weaksrs strictsrs
      putStrLn p
   )         
   