module Main where
import Data.List
import Data.Char
import CYCSRS.CallsParam
import CYCSRS.Parser
import CYCSRS.Processor
import CYCSRS.Processors.AProVE
import CYCSRS.Processors.TTT2
import CYCSRS.Processors.Combinations
import CYCSRS.Processors.Encoders.Rotate
import CYCSRS.Processors.Encoders.Shift
import CYCSRS.Processors.Encoders.Split
import CYCSRS.Processors.Encoders.SRS2CYC
import CYCSRS.Processors.Encoders.Tools
import CYCSRS.Processors.MatrixInterpretator
import CYCSRS.Processors.Nontermination
import CYCSRS.Processors.Cycnt
import CYCSRS.Processors.Torpacyc
import CYCSRS.Types
import System.Environment
import CYCSRS.Util.System(deleteTMPFiles)
type Timeout = Int

main = 
 do 
  args <- getArgs
  case parseOptions args of
    Right options -> case options of 
                       HELP -> displayHelp 
                       opts -> do 
                               runProc (relative opts) (combine_sequence (processors opts)) (input_file opts) (timeoutval opts)
                               deleteTMPFiles
    Left err -> displayErr err

internalMain  opts = 
   do 
     res <- runProcString (relative opts) (combine_sequence (processors opts)) (input_file opts) (timeoutval opts)
     deleteTMPFiles
     return res
    
displayErr err =
  do
    name <- getProgName
    putStrLn $ "error: " ++ err ++ "\ntype '"++ name++ " help' to display all options" 

displayHelp = 
   do 
     name <-getProgName
     putStrLn $ helptext name
     



data Options = Options {input_file :: FilePath,
                        processors :: [(Processor,Ratio)],
                        timeoutval :: Timeout,
                        relative :: Bool}
              | HELP
instance Show Options where
  show options = 
    concat $ intersperse " " [input_file options] ++ (map show $ map snd $ processors options) ++ [show $ timeoutval options]

    
    
parseOptions ["help"] = Right HELP    
parseOptions ["-h"] = Right HELP    
parseOptions options
  | length options < 3 = Left "unsufficient number of arguments"
  | otherwise          =
     let (timeoutarg:infile:reversed_options) =  reverse options
         processorargs       = reverse reversed_options
         processorargs'      = delete "relative" processorargs
         isRelative          = case processorargs of 
                                ("relative":_) -> True         
                                _ -> False
     in if all isDigit timeoutarg then
          case (parseOptions2 processorargs') of
           Left err -> Left err
           Right procs -> if (sum (map snd procs)) == 100
                           then Right $ Options {processors=procs,input_file=infile,timeoutval = read timeoutarg,relative=isRelative}
                           else Left $ "the sum of all ratio values must be 100, but it is " ++
                                                                    ((concat $ intersperse "+" (map show $ map snd procs))) ++ "=" 
                                                                      ++ (show $sum $ map snd procs)
         else 
           Left $ "Timeout " ++ show timeoutarg ++ " not valid, must be a non-negative integer"

parseOptions2 []        = Right []          
parseOptions2 [p]       = case parseProc p of
                            Left err -> case parseProcessor p of
                                          Left err -> Left err
                                          Right proc -> Right [(proc,100)]
                            Right x -> Right [x]
parseOptions2 (p:procs) =
  case parseProc p of
    Left err -> Left err 
    Right p' -> case parseOptions2 procs of            
                  Left err -> Left err
                  Right ps' -> Right (p':ps')
  
  
parseProc :: String -> Either String (Processor,Timeout)  
parseProc (p) = 
  case break (== '-') p of
    (proc,[]) -> Left $ "missing ratio  after processor : " ++ show p
    (proc,'-':rest) ->  
                     if all isDigit rest then
                      case parseProcessor proc of
                        Left err    -> Left err
                        Right proc' -> Right (proc',read rest) 
                     else 
                       Left $ "failed to parse ratio " ++ show rest ++ " in processor " ++ show p
    _ -> Left $ "failed to parse processor" ++ show p                  
                         
parseProcessor :: String -> Either String Processor
parseProcessor str =
 case lookup str available_processors of
   Nothing -> Left $ "processor " ++ str ++ " is unknown"
   Just p  -> Right p
   
   
available_processors =
  [
   ("aprove",         CYCSRS.Processors.AProVE.processor)
  ,("ttt2",           CYCSRS.Processors.TTT2.processor)
  ,("split",          CYCSRS.Processors.Encoders.Split.processor)
  ,("shift",          CYCSRS.Processors.Encoders.Shift.processor)
  ,("split_aprove",   CYCSRS.Processors.Combinations.transform_and_aprove CYCSRS.Processors.Encoders.Split.processor)
  ,("shift_aprove",   CYCSRS.Processors.Combinations.transform_and_aprove CYCSRS.Processors.Encoders.Shift.processor)
  ,("rotate_aprove",  CYCSRS.Processors.Combinations.transform_and_aprove CYCSRS.Processors.Encoders.Rotate.processor)
  ,("split_ttt2",     CYCSRS.Processors.Combinations.transform_and_ttt2 CYCSRS.Processors.Encoders.Split.processor)
  ,("shift_ttt2",     CYCSRS.Processors.Combinations.transform_and_ttt2 CYCSRS.Processors.Encoders.Shift.processor)
  ,("rotate_ttt2",    CYCSRS.Processors.Combinations.transform_and_ttt2 CYCSRS.Processors.Encoders.Rotate.processor)
  ,("rotate",         CYCSRS.Processors.Encoders.Rotate.processor)
  ,("torpacyc",       CYCSRS.Processors.Torpacyc.processor)
  ,("tdmi",           CYCSRS.Processors.MatrixInterpretator.processor)
  ,("cycnt",          CYCSRS.Processors.Cycnt.processor)
  ,("default",        CYCSRS.Processors.Combinations.default_processor)
  ,("default_long",   CYCSRS.Processors.Combinations.default_long_processor)
  ]
  
  


helptext name =
 unlines $
 [
  "AVAILABLE OPTIONS"
 ," * "++ name ++" [relative] PROCESSORS INFILE TIMEOUT"
 ,"   Apply the processors sequentially to string rewrite system in INFILE"
 ," * "++ name ++" help"
 ,"   Show help"
 ,""
 ,"      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
 ,"      !!! ENSURE that the NECESSARY SOFTWARE is installed and           !!!"
 ,"      !!! ADJUST the pathes to your LOCAL CONFIGURATION in the file     !!!"
 ,"      !!!                                                               !!!"
 ,"      !!!        conf.sh                                                !!!"
 ,"      !!! Note that cycsrs will create a default conf.sh                !!!"
 ,"      !!! if it is needed and not available                             !!!"
 ,"      !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
 ,""
 ,""
 ," [relative] (optional)"
 ,"   if this parameter is given, then relative termination instead" 
 ,"   of termination is proved. The input must be in xml-Format."
 ,""
 ,""
 ," PROCESSORS"
 ," the general format is "
 ,"     proc1-ratio1 proc2-ratio2 proc3-ratio3 ..."
 ," where"
 ,"   procI is the name of a processor"
 ,"   ratioI is the portion in percent of the timeout, given to processor I"
 ,"   the sum ratio1+ratio2+ratio3+...  must be exactly 100"
 ,"   if a single processor is used also proc instead of proc-100 is allowed"
 ," the processors are run in sequence proc1 proc2 proc3 ..."
 ," where the timeout for each processor is adjusted accordingly"
 ," available processors are:"
 ,"  torpacyc      = proves cycle termination by matrix interpretations"
 ,"  aprove        = termination prover AProVE applied to the"
 ,"                  string rewrite system. "
 ,"                  This prover cannot show cycle-termination, but it can" 
 ,"                  infer cycle-nontermation by proving string-nontermination"
 ,"  ttt2          = termination prover TTT2 applied to the"
 ,"                  string rewrite system. "
 ,"                  This prover cannot show cycle-termination, but it can" 
 ,"                  infer cycle-nontermation by proving string-nontermination"
 ,"  split_aprove  = apply the sound and complete transformation split (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  rotate_aprove = apply the sound and complete transformation rotate (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  shift_aprove  = apply the sound and complete transformation shift (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  split_ttt2    = apply the sound and complete transformation split (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  rotate_ttt2   = apply the sound and complete transformation rotate (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  shift_ttt2    = apply the sound and complete transformation shift (SZ15)"
 ,"                  to transform the cycle-termination problem in"
 ,"                  a string termination problem and then apply the ATP"
 ,"                  AProVE to solve the string termination problem"
 ,"  tdmi          = searches for cycle-termination by trace-decreasing"
 ,"                  matrix interpretation (SZ15) using the SMT-Solver" 
 ,"                  Yices2."
 ,"  split         = apply the sound and complete transformation split (SZ15)"
 ,"  rotate        = apply the sound and complete transformation rotate (SZ15)"
 ,"  shift         = apply the sound and complete transformation shift (SZ15)"
 ,"  cycnt         = a brute-force search for cycle-nontermination"
 ,"  default       = is equivalent to torpacyc-40 cycnt-10 aprove-6 split_aprove 42"
 ,"      "
 ,"INFILE"
 ," the format of the content is resolved by its file extension"
 ,"  .srs   = string rewrite system in text format"
 ,"  .xml   = string rewrite system in XML format of the TPDB"
 ,"      " 
 ,"TIMEOUT"
 ," a number of seconds"
 ,"      "
 ,"Configuration"
 ,"============="
 ," the available parameters can be adjusted in the file"
 ,"    conf.sh "
 ,""
 ]
                
