module TDMI.Generic where          
import Data.Maybe
import Data.List
import Debug.Trace
import qualified CYCSRS.Types as T

-- =============================================
-- symbolic representation of arithmetic expressions including Exoctic Semirings and conditions

data ArithEx = Var String | ArithEx :*: ArithEx | ArithEx :+: ArithEx | Number Integer | MinInfty | MaxInfty
 deriving (Show)
 
data Cond = ArithEx :>: ArithEx | ArithEx :>=: ArithEx | ArithEx :=: ArithEx | NOT Cond | AND [Cond] | OR [Cond] | EQUIV Cond Cond
    deriving (Show)

 
-- small help functions:
isVar (Var _) = True
isVar _ = False
fromVar (Var v) = v
 
-- =============================================


instance Show Domain where
     show d = domainName d
data Domain = 
 Domain {
     domainName :: String
    ,zero :: ArithEx
    ,one :: ArithEx
    ,domainToInteger :: ArithEx -> Integer
    ,intMult :: Integer -> Integer -> Integer 
    ,intPlus :: Integer -> Integer -> Integer
    ,lessthan :: Integer -> Integer -> Bool
    ,equals :: Integer -> Integer -> Bool
    ,leq :: Integer -> Integer -> Bool
    ,showNum :: Integer -> String
-- fixBitWidth receives
--  - bitwidth the true bitwidth used for bit-vectors
--  - (conditions,alphabet) the conditions of trace-decreasing matrix interpretations and
--    the alphabet, i.e. the variables representing numbers
--  - bound: the bitwidth which should not be exceeded by the interpretation of the symbols
--    note that bound is usually much smaller than bitwidth, since the arithmetic operations
--    increase the numbers and bitwidth must be large enough to represent all these numbers
-- the output of fixBitWidth is a string representing the SMT-Problem (using BitVecs) as STM-LIB2
-- problem. Additionally to generating the formulas for the condition, it generates formulas
-- that the numbers of the interpretation are small enough to be representable in the bit width.
     ,fixBitWidth :: Int -> (Cond, [T.Symbol]) -> Int -> String
     ,strictsmaller :: (Matrix ArithEx) -> (Matrix ArithEx) -> Int -> Cond    
     ,strictlySmaller :: Matrix Integer -> Matrix Integer -> Bool
     ,weaklySmaller :: Matrix Integer -> Matrix Integer -> Bool
     ,conditionSymbol :: Int -> [[ArithEx]] -> Cond

 }
 
 
 
-- =============================================
-- Square Matrices

type Matrix a = [[a]]
-- simple access
xs !!! (x,y)  = (xs!!(y-1))!!(x-1) 

-- special matrices
-- identity matrix for ArithEx
idm domain dim = 
 [[ if col == row then (one domain) else (zero domain) | col <- [1..dim]] | row <- [1..dim]]
-- identity matrix for Integer 
idmInt domain dim = 
 [[ if col == row then  (domainToInteger domain (one domain)) else (domainToInteger domain (zero domain)) | col <- [1..dim]] | row <- [1..dim]]

-- make matrix (with symbolic variables as entries) for a symbol (sym) given of a specific dimension (dim)
-- the representation for one entry in the matrix of symbol "abc" is  
-- "a_row_col_abc": 
--    the 'a' in front of the variable is necessary to deal with symbols that start
--    with a special letter (e.g. number), otherwise yices cannot handle these symbols
interpretation dim sym = 
   [[ Var $ "a_" ++ show row ++ "_" ++ show col ++ "_" ++ sym | col <- [1..dim]] | row <- [1..dim]]
-- =============================================


-- =============================================
-- matrix multiplication

-- generic version
mmult m1 m2 dim times plus =
   [[ vec times plus (getRow m1 row) (getCol m2 col) | col <- [1..dim]] | row <- [1..dim]]
  where
     vec times plus xs ys = foldl1 plus $ zipWith times xs ys
     getRow m i = m!!(i-1)
     getCol m i = getRow (transpose m) i

-- symbolic matrix multiplication on ArithEx
mxmult m1 m2 dim = mmult m1 m2 dim (:*:) (:+:)
-- =============================================


-- =============================================
-- Conditions for trace-decreasing matrix interpretation (SZ15-RTA)

-- makeConditions generates all conditions for a trace-decreasing matrix interpretation:
--  (a) every symbol must be interpreted as a square matrix of M_D where D is the dimension
--      an M_D is the set of all square matrices of dimension D where all entries are 
--      non-negative integers and the upper-left entry is >= 1
--      [these conditions are generated by conditionsOnMatrices]
--  (b) all rules are weakly decreasing and at least one rule is strictly decreasing
--      [these conditions are generated by termcond]
-- the output of makeConditions is the Condition and the alphabet (for further processing)
makeConditions  ::  Bool -> Domain -> Int -> (T.SRS,T.SRS) -> (Cond,T.Alphabet)
makeConditions False domain dim (_,srs) = 
 let alphabet = T.getAlphabet srs 
     allMatrices = (map (interpretation dim) (alphabet)) 
     symbols = map fromVar $ filter isVar $ concat $ concat allMatrices
     conditionsOnMatrices = AND $ map (conditionSymbol domain dim) allMatrices
     l = length srs
     termcond = if l >= 2 then 
        OR [AND ((conditionRuleStrict domain  dim (srs!!j)):[conditionRuleWeak domain dim (srs!!i) | i <- [0..l-1], i /= j]) | j <- [0..l-1]]
                else if l == 1 then conditionRuleStrict domain dim (head srs) else error "too few rules"
 in (AND [conditionsOnMatrices,termcond], symbols)
-- relative termination:
makeConditions True domain dim (weaksrs,strictsrs) = 
 let srs = strictsrs ++ weaksrs
     alphabet = T.getAlphabet srs 
     allMatrices = (map (interpretation dim) (alphabet)) 
     symbols = map fromVar $ filter isVar $ concat $ concat allMatrices
     conditionsOnMatrices = AND $ map (conditionSymbol domain dim) allMatrices
     l = length srs
     ls = length strictsrs
     lw = length weaksrs
     wcond = [conditionRuleWeak domain dim rule | rule <- weaksrs]
     termcond = if l >= 2 then 
        OR [AND (((conditionRuleStrict domain dim (srs!!j)):[conditionRuleWeak domain dim (srs!!i) | i <- [0..ls-1], i /= j]) ++ wcond) | j <- [0..ls-1]]
                else if l == 1 then conditionRuleStrict domain dim (head srs) else error "too few rules"
 in (AND [conditionsOnMatrices,termcond], symbols)

 
  
-- conditions for string rewrite rules (l,r)
-- (1) the rule is weakly decreasing, i.e. [[l]] >= [[r]]

conditionRuleWeak domain dim (l,r)  =
   let thismult a b = mxmult a b dim
       left = foldl  thismult (idm domain dim) (map (interpretation dim) l)
       right = foldl thismult (idm domain dim) (map (interpretation dim) r)
    in weaksmaller left right dim
 where
   -- weaksmaller: all entries are >=
   weaksmaller :: (Matrix ArithEx) -> (Matrix ArithEx) -> Int -> Cond    
   weaksmaller m1 m2 dim =  AND [(m1!!!(row,col)) :>=: (m2!!!(row,col)) | row <- [1..dim], col <- [1..dim]] 

-- (2) or the rule is strictly decreasing, i.e. [[l]] > [[r]]
   
conditionRuleStrict domain dim (l,r)  =
   let thismult a b = mxmult a b dim
       left = foldl thismult (idm domain dim) (map (interpretation dim) l)
       right = foldl thismult (idm domain dim) (map (interpretation dim) r) 
   in strictsmaller domain left right dim
-- =============================================
  
          
printSetOfRulesRel rules = unlines (map printOneRuleRel rules)
printSetOfRules rules = unlines (map printOneRule rules)
printOneRuleRel (num,(a,b),(ml,mr)) =
  "rule (" ++ show num ++ ") " ++ T.showRuleRel (a,b) 
printOneRule (num,(a,b),(ml,mr)) =
  "rule (" ++ show num ++ ") " ++ T.showRule (a,b) 

          
printCheckL domain xs = concat $ intersperse "\n\n" $ map (printCheck domain) xs          
printCheckLRel domain xs = concat $ intersperse "\n\n" $ map (printCheckRel domain) xs          
printCheckRel domain (num,r@(a,b),(ml,mr)) =
  unlines
  ["(" ++ show num ++ ") " ++ T.showRuleRel r, "Interpretation of the lhs " ++ T.showSequence a ++ "\n", printMatrix domain ml, "Interpretation of the rhs " ++ T.showSequence b ++ "\n",printMatrix domain mr]
printCheck domain (num,r@(a,b),(ml,mr)) =
  unlines
  ["(" ++ show num ++ ") " ++ T.showRule r, "Interpretation of the lhs " ++ T.showSequence a ++ "\n", printMatrix domain ml, "Interpretation of the rhs " ++ T.showSequence b ++ "\n",printMatrix domain mr]
  
-- strictrule domain inp@(num,r@(a,b),(ml,mr))
--   | trace ("strictrule?:" ++ show inp ++ " out:" ++ (show  $ strictlySmaller domain mr ml)) False = undefined
strictrule domain inp@(num,r@(a,b),(ml,mr)) =
  strictlySmaller domain mr ml

   

check_interpretation_of_rules i domain dim srs interpret = map (\(a,r) -> (a,r,check_interpretation_of_a_rule domain dim r interpret)) (zip [i..] srs)

check_interpretation_of_a_rule domain dim (l,r) interpret =
   let thismult a b = mmult a b dim (intMult domain) (intPlus domain)
       interpretS a = fromJust $  lookup a interpret  
       left =  foldl thismult (idmInt domain dim) (map interpretS l)
       right = foldl thismult (idmInt domain dim) (map interpretS r)
   in 
       
       (left,right)
   
  
 



parseYicesOutput xs = 
   let tokens = tokenize xs
   in grouping $ runAll tokens
 where
  -- tokenize
  tokenize [] = []
  tokenize (s:xs) 
    | s `elem` ['(',')'] = [s]:(tokenize xs)
  tokenize ('a':'_':xs) =
    let (var,rest) = break (\x -> x == ' ' || x == '\n') xs
    in (var:(tokenize rest))
  tokenize ('#':'b':xs) = 
    let (r,rs) = span (\x -> x `elem` ['0','1']) xs 
    in r:(tokenize rs)
  tokenize (_:xs) = (tokenize xs)
  -- runAll
  runAll xs = if head xs == "(" && last xs == ")"
               then run_inner (reverse $ tail $ reverse $ tail xs)
	       else error $ show xs
  -- run_inner
  run_inner ("(":var:b:")":rest) = (parseInt var (bv2int (reverse b) 0)):(run_inner rest)
  run_inner [] = []
  run_inner xs = error $ show xs
  -- bv2int
  bv2int ('0':xs) j = bv2int xs (j+1)
  bv2int ('1':xs) j = 2^j + (bv2int xs (j+1))
  bv2int [] _ = 0
  -- parseInt
  parseInt :: String -> Integer -> (String,(Int,Int),Integer)
  parseInt xs i =
    let 
      (x,'_':rest) = break (== '_') xs
      (y,'_':var)  = break (== '_') rest
    in (var,((read x),(read y)),i)
  -- grouping
  grouping xs = map toMatrix $ map transform $ groupBy cmp xs
    where cmp (a,b,c) (d,e,f) = a == d
          transform (xs@((a,_,_):_)) = (a,map (\(a,b,c) -> (b,c)) xs)
          toMatrix (a,xs) = let (i,j) = maximum $ map fst xs
                            in (a,[[ fromJust $ lookup (row,col) xs | col <- [1..j]] | row <- [1..i]])
                            
printAllInterpret domain xs = concat $ intersperse "\n\n" $  map (printInterpret domain) xs			  
printInterpret domain (a,xs) = (a ++ " |-> ") ++ "\n" ++ (unlines (map (printLine domain) xs))
printLine domain xs = "            |" ++ (concat $ intersperse " " (map (showNum domain) xs)) ++ "|"
printMatrix domain xs = unlines $  map (printLine domain) xs			  


-- helper functions for appending and prepending to the proof
-- todo: maybe placed in another module, and the Implementation maybe more efficient
  